package org.globus.transfer.reliable.service.client;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.wsrf.ResourceProperty;
import org.globus.ftp.MlsxEntry;
import org.globus.rft.generated.TotalBytesTransferredResource;
import org.globus.transfer.reliable.service.MyMarkerListener;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.transfer.reliable.service.UserSubjectUtil;
import org.globus.transfer.reliable.service.cache.ConnectionManager;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.cache.ThirdPartyConnectionImpl;
import org.globus.transfer.reliable.service.cache.ConnectionParameter;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.ietf.jgss.GSSCredential;

public class ThirdPartyTransferClient implements Client
{   
    private static Log logger = 
        LogFactory.getLog(ThirdPartyTransferClient.class.getName());

    private long size;
    private int transferId;

    private String restartMarker;
    private String permissions;

    private ConnectionManager connectionManager = null;
    private ThirdPartyConnectionImpl connection = null;
    private ResourceProperty totalBytesTransferredRP;
    private TotalBytesTransferredResource totalBytes;
    
    public ThirdPartyTransferClient(
            ConnectionParameter  params,
            GSSCredential        credential,
            ConnectionManager    connectionManager,
            ResourcePropertySet  rpSet,
            int                  transferId)
    throws RftException, RftTransientException, RftDBException
    {
    	if (params == null) {
            throw new RftException("parameter for connection " +
                "must not be null");
        }
        if (credential == null) {
        	throw new RftException("credential must not be null");
        }
        if (connectionManager == null) {
            throw new RftException("connectionManager must not be null");
        }
        if (rpSet == null) {
            throw new RftException("ResourcePropertySet " +
                "must not be null");
        }



    	logger.debug("getting connection from the connection manager ("+
        		    ConnectionManager.THIRDPARTY_CONNECTION + ")");
        // get a connection from the ConnectionManager
        this.connectionManager = connectionManager;
        this.connection = (ThirdPartyConnectionImpl)
            this.connectionManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION,
                params,
                credential,
                UserSubjectUtil.getSourceUserSubject(params.getRFTOptions()),
                UserSubjectUtil.getDestinationUserSubject(params.getRFTOptions()));
        
        this.transferId = transferId;
        this.totalBytesTransferredRP
               = rpSet.get(RFTConstants.TOTAL_BYTES_RESOURCE);
        this.totalBytes
               = (TotalBytesTransferredResource)
                    totalBytesTransferredRP.get(0);

        // get restart marker from database
        logger.debug("getting restart marker");
        this.restartMarker = ReliableFileTransferDbAdapter
                                   .getDbAdapter()
                                      .getRestartMarker(transferId);
    }
    
    /**
     * 
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public void process()
    throws RftException, RftTransientException, RftDBException
    {
        boolean success = false;
        try
        {
            // get permissions and size from the source
            this.doMlstOnSource();
        
            // perform the transfer
            this.transfer();
        
            // set permissions on the destination
            connection.setPermissionsOnDestination(this.permissions);
    
            // update total bytes
            this.totalBytes.setTotalBytes(
                    this.totalBytes.getTotalBytes() + this.size);
            this.totalBytesTransferredRP.set(0, this.totalBytes);
        
            // update last used and add connection back to the cache
            connection.updateLastUsed();
            this.connectionManager.addConnection(connection);
            success = true;
        }
        finally
        {
            if(!success)
            {
                connection.closeConnection();
            }
            else
            {
                connection.successfulTransfer();
            }
        }
    }
    
    /**
     * Performs the actual transfer
     * 
     * @param useExtended
     * @throws RftTransientException If setting the restart markers fails or
     *                               if the transfer fails
     */
    private void transfer() throws RftTransientException
    {
        MyMarkerListener markerListener = new MyMarkerListener(transferId);
        if (this.restartMarker != null) {
        	connection.setRestartMarker(this.restartMarker);
        	connection.normalExtendedTransfer(markerListener);
        } else {
            connection.normalNonExtendedTransfer(markerListener);
        }
    }

    /**
     * This method gets the permissions of the directory on the
     * server-side and also the size. Both are stored in database.
     * @return
     * @throws RftException If the file/dir on the source does not exist or
     *                      if the the interaction with the DB system fails.
     * @throws RftTransientException If any other error occurs during GridFTP
     *                               server interaction
     * @throws RftDBException
     */
    private void doMlstOnSource() throws RftException,
        RftTransientException, RftDBException
    {
        MlsxEntry entry = connection.doMlstOnSource();
        this.size = Long.parseLong(entry.get(MlsxEntry.SIZE));
        this.permissions = entry.get("unix.mode");

        ReliableFileTransferDbAdapter.getDbAdapter().updateTransfer(
            this.transferId, this.size, this.permissions);
    }
}
